#ifndef PLUGINCLIENT_H
#define PLUGINCLIENT_H

// Base class inherited by all the different types of plugins.

#define BCASTDIR "~/.bcast/"

class PluginClient;


#include "arraylist.h"
#include "keyframe.h"
#include "mainprogress.inc"
#include "maxbuffers.h"
#include "messages.h"
#include "pluginbuffer.h"
#include "plugincommands.h"
#include "pluginserver.inc"
#include "sema.h"
#include "theme.inc"
#include "vframe.inc"


extern "C"
{
extern PluginClient* new_plugin(PluginServer *server);
}

class PluginClientAuto
{
public:
	float position;
	float intercept;
	float slope;
};

class PluginClient
{
public:
	PluginClient(PluginServer *server);
	virtual ~PluginClient();

	
// Queries for the plugin server.
	virtual int is_realtime();
	virtual int is_audio();
	virtual int is_video();
	virtual int is_fileio();
	virtual int is_theme();
	virtual int uses_gui();
	virtual int is_multichannel();
	virtual int is_synthesis();
	virtual int is_transition();
	virtual char* plugin_title();   // return the title of the plugin
	virtual VFrame* new_picon();
	virtual Theme* new_theme();






// Non realtime signal processors define these.
	virtual int get_samplerate();        // give the samplerate of the output for a non realtime plugin
	virtual float get_framerate();         // give the framerate of the output for a non realtime plugin
	virtual int init_nonrealtime_parameters();   // Set up pointers for plugin
	virtual int delete_nonrealtime_parameters();
	virtual int start_plugin();         // run a non realtime plugin
	virtual int get_parameters();     // get information from user before non realtime processing
	virtual long get_in_buffers(long recommended_size);  // return desired size for input buffers
	virtual long get_out_buffers(long recommended_size);     // return desired size for output buffers
	virtual int start_loop();
	virtual int process_loop();
	virtual int stop_loop();




// Realtime commands for signal processors.
// These must be defined by the plugin itself.
	virtual int set_string();             // Set the string identifying the plugin to modules and patches.
// cause the plugin to show the gui
	virtual int show_gui();               
// cause the plugin to hide the gui
	void client_side_close();
	void update_display_title();
// Raise the GUI
	virtual void raise_window() {};
	virtual int start_realtime();         // start the plugin waiting to process realtime data
	virtual int stop_realtime();          // start the plugin waiting to process realtime data
	virtual void save_data(KeyFrame *keyframe) {};    // write the plugin settings to text in text format
	virtual void read_data(KeyFrame *keyframe) {};    // read the plugin settings from the text
// Return keyframe objects.  If position -1 use edl selection
	KeyFrame* get_prev_keyframe(long position);
	KeyFrame* get_next_keyframe(long position);
	int send_hide_gui();                                    // should be sent when the GUI recieves a close event from the user
	int send_configure_change();                            // when this plugin is adjusted, propogate parameters to virtual plugins
	int get_configure_change();                             // get propogated configuration change from a send_configure_change
	int plugin_restart_realtime();        // Get new buffers after a rebuilt virtual console
	virtual void plugin_process_realtime(double **input, 
		double **output, 
		long current_position, 
		long fragment_size,
		long total_len) {};
	virtual void plugin_process_realtime(VFrame **input, 
		VFrame **output, 
		long current_position,
		long total_len) {};
	virtual int plugin_process_loop(VFrame **buffers, long &write_length) { return 1; };
	virtual int plugin_process_loop(double **buffers, long &write_length) { return 1; };
	virtual int init_realtime_parameters();     // get parameters depending on video or audio
	int get_gui_status();
// Realtime parameters for transient plugins which are called during process_realtime.
	long get_source_len();
	long get_source_position();
// If automation is used
	int automation_used();
// Get the automation value for the position in the current fragment
// The position is relative to the current fragment
	float get_automation_value(long position);




// Operations for file handlers
	virtual int open_file() { return 0; };
	virtual int get_audio_parameters() { return 0; };
	virtual int get_video_parameters() { return 0; };
	virtual int check_header(char *path) { return 0; };
	virtual int open_file(char *path, int wr, int rd) { return 1; };
	virtual int close_file() { return 0; };





// All plugins define these.
	virtual int load_defaults();       // load default settings for the plugin
	virtual int save_defaults();      // save the current settings as defaults




// Non realtime operations for signal processors.
	int plugin_start_loop(long start, long end, long buffer_size, int total_buffers);
	int plugin_stop_loop();
	int plugin_process_loop();
	MainProgressBar* start_progress(char *string, long length);
	int get_project_samplerate();            // get samplerate of project data before processing
	float get_project_framerate();         // get framerate of project data before processing
	int get_gui_visible();     // Get the GUI status from the server before processing
	int get_project_framesize(int &frame_w, int &frame_h); // get frame size of project data before processing
	int get_use_float();
	int get_use_alpha();
	int get_use_interpolation();
	int get_project_smp();
	int get_aspect_ratio(float &aspect_w, float &aspect_h);
	int write_frames(long total_frames);  // returns 1 for failure / tells the server that all output channel buffers are ready to go
	int write_samples(long total_samples);  // returns 1 for failure / tells the server that all output channel buffers are ready to go
	int plugin_get_parameters();
	char* get_defaultdir();     // Directory defaults should be stored in
	int plugin_get_samplerate();
	float plugin_get_framerate();
	void set_interactive();

// Realtime operations.
	int plugin_init(int argc, char *argv[]);
	int reset();
	int plugin_run();
	virtual int plugin_command_derived(int plugin_command) {}; // Extension of plugin_run for derived plugins
	int plugin_get_range();
	int plugin_negotiate_buffers();
	int plugin_start_plugin();    // Run a non realtime plugin
	int plugin_init_realtime(int realtime_priority, int total_in_buffers);
	int plugin_stop_realtime();
	int plugin_create_buffers();
	int plugin_delete_buffers();



// create pointers to buffers of the plugin's type before realtime rendering
	virtual int create_buffer_ptrs();
	virtual int delete_buffer_ptrs();




// communication convenience routines for the base class
	int stop_gui_client();     
	int save_data_client();
	int load_data_client();
	int set_string_client();                // set the string identifying the plugin
	int send_completed();        // send when finished
	int send_cancelled();        // non realtime plugin sends when cancelled

// ================================= Buffers ===============================
// non realtime buffers
	PluginBuffer **data_in;
	PluginBuffer **data_out;
// realtime buffers
	ArrayList<PluginBuffer**> data_in_realtime;   // data coming in from plugin
	ArrayList<PluginBuffer**> data_out_realtime;  // data going out to plugin

// number of double buffers for each channel
	ArrayList<int> double_buffers_in;    
	ArrayList<int> double_buffers_out;
// When arming buffers need to know the offsets in all the buffers and which
// double buffers for each channel before rendering.
	ArrayList<long> offset_in_render;
	ArrayList<long> offset_out_render;
	ArrayList<long> double_buffer_in_render;
	ArrayList<long> double_buffer_out_render;
// total size of each buffer depends on if it's a master or node
	ArrayList<long> realtime_in_size;
	ArrayList<long> realtime_out_size;

// ================================= Automation ===========================

	ArrayList<PluginClientAuto> automation;

// ================================== Messages ===========================
	Messages *messages, *gui_messages;
	Sema *message_lock;             // only used for realtime plugins
	char gui_string[1024];          // string identifying module and plugin
	int master_gui_on;              // Status of the master gui plugin
	int client_gui_on;              // Status of this client's gui

	int show_initially;             // set to show a realtime plugin initially
// range in project for processing
	long start, end;                
	int interactive;                // for the progress bar plugin
	int success;
	int total_out_buffers;          // total send buffers allocated by the server
	int total_in_buffers;           // total recieve buffers allocated by the server
	int wr, rd;                     // File permissions for fileio plugins.

// these give the largest fragment the plugin is expected to handle
// currently these give the master size
	long out_buffer_size;  // size of a send buffer to the server
	long in_buffer_size;   // size of a recieve buffer from the server
// Local parameters of plugin
	int sample_rate;
	double frame_rate;
// Operating system scheduling
	int realtime_priority;

// Realtime parameters for transient plugins.
	long source_len;
	long source_position;
	long total_len;
// Total number of processors available
	int smp;  
	PluginServer *server;

private:

// File handlers:
//	Asset *asset;     // Point to asset structure in shared memory
};


#endif
