#include "assets.h"
#include "audiodevice.h"
#include "bcdisplayinfo.h"
#include "defaults.h"
#include "filemov.h"
#include "montevideo.h"
#include "mvaudioin.h"
#include "mvaudioout.h"
#include "mvconsole.h"
#include "mvdatabase.h"
#include "mvinterface.h"
#include "mvvideoin.h"
#include "mvvideoout.h"
#include "mvwindow.h"
#include "mwindow.h"
#include "playbackconfig.h"
#include "quicktime.h"
#include "recordconfig.h"
#include "vframe.h"
#include "videodevice.h"

#include <string.h>

int main(int argc, char *argv[])
{
	MonteVideo main_loop;
	main_loop.create_objects(argc, argv);
	main_loop.main_loop();
}


MonteVideo::MonteVideo()
{
	reset();
}

MonteVideo::~MonteVideo()
{
	delete interface;
	if(window) delete window;
	if(console) delete console;
	delete overlay;
}

void MonteVideo::reset()
{
	done = 0;
	aconfig_in = 0;
	vconfig_in = 0;
	video_in = 0;
	audio_in = 0;
	aconfig_out = 0;
	vconfig_out = 0;
	video_out = 0;
	audio_in = 0;
	audio_out = 0;
	asset = 0;
	cpus = 1;
	window = 0;
	console = 0;
	state = LOOPBACK;
	overlay_start = overlay_end = overlay_eternal = 0;
}

void MonteVideo::usage()
{
	printf("\nMontevideo: Unleash the VCR in your web server\n"
		"-p path           Prefix for data files.  (%s)\n"
		"-g wwwxhhh        Working frame size (%dx%d)\n"
		"-r sample rate    Working sample rate  (%d)\n"
		"-f frame rate     Target frame rate  (%.2f)\n"
		"-c compressor     Compression format (%s)\n"
		"        0) %s\n"
		"        1) %s\n"
//		"        2) %s\n"
//		"        3) %s\n"
		"-q quality        JPEG quality 1-100 (%d)\n"
		"\n"
		"-O path           Path of output device (%s)\n"
		"-o device ID      Output driver: (%s)\n"
		"        0) %s\n"
		"        1) %s\n"
		"\n"
		"-I path           Path of input device (%s)\n"
		"-i device ID      Input driver: (%s)\n"
		"        0) %s\n"
		"        1) %s\n"
		"        2) %s\n", 
		asset->path, 
		asset->width,
		asset->height,
		asset->rate, 
		asset->frame_rate,
		FileMOV::compressiontostr(asset->compression),
		inttocompressor(0),
		inttocompressor(1),
//		inttocompressor(2),
//		inttocompressor(3),
		asset->quality,
		vconfig_out->get_path(),
		VideoDevice::drivertostr(vconfig_out->driver), 
		VideoDevice::drivertostr(PLAYBACK_X11), 
		VideoDevice::drivertostr(PLAYBACK_BUZ),
		vconfig_in->get_path(), 
		VideoDevice::drivertostr(vconfig_in->driver), 
		VideoDevice::drivertostr(VIDEO4LINUX), 
		VideoDevice::drivertostr(SCREENCAPTURE), 
		VideoDevice::drivertostr(CAPTURE_BUZ));
}

char* MonteVideo::inttofourcc(int compressor)
{
	switch(compressor)
	{
		case 0:
			return QUICKTIME_JPEG;
			break;
		case 1:
			return QUICKTIME_MJPA;
			break;
		case 2:
			return QUICKTIME_YUV420;
			break;
		case 3:
			return QUICKTIME_YUV2;
			break;
	}
	return QUICKTIME_JPEG;
}

char* MonteVideo::inttocompressor(int compressor)
{
	switch(compressor)
	{
		case 0:
			return "JPEG Photo";
			break;
		case 1:
			return "Motion JPEG A";
			break;
		case 2:
			return "YUV 4:2:0";
			break;
		case 3:
			return "YUV 4:2:2";
			break;
	}
	return "Unknown";
}

// Some things must be specified on the command line
void MonteVideo::parse_args(int argc, char *argv[])
{
	int n;
	while((n = getopt(argc, argv, "g:r:f:o:q:O:i:I:p")) != EOF)
	{
		switch(n)
		{
			case 'g':
			{
				int temp_x, temp_y;
				BC_DisplayInfo::parse_geometry(optarg, 
					&temp_x, 
					&temp_y, 
					&asset->width, 
					&asset->height);
				break;
			}

			case 'r':
				asset->rate = atol(optarg);
				break;

			case 'f':
				asset->frame_rate = atof(optarg);
				break;

			case 'c':
				strcpy(asset->compression, inttofourcc(atol(optarg)));
				break;

			case 'o':
				vconfig_out->driver = atoi(optarg);
				switch(vconfig_out->driver)
				{
					case 0:
						vconfig_out->driver = PLAYBACK_X11;
						break;
					case 1:
						vconfig_out->driver = PLAYBACK_BUZ;
						break;
				}
				break;
			
			case 'q':
				asset->quality = atol(optarg);
				break;

			case 'O':
				strcpy(vconfig_out->get_path(), optarg);
				break;

			case 'i':
				vconfig_in->driver = atoi(optarg);
				switch(vconfig_in->driver)
				{
					case 0:
						vconfig_in->driver = VIDEO4LINUX;
						break;
					case 1:
						vconfig_in->driver = SCREENCAPTURE;
						break;
					case 2:
						vconfig_in->driver = CAPTURE_BUZ;
						break;
				}
				break;

			case 'I':
				strcpy(vconfig_in->get_path(), optarg);
				break;

			case 'p':
				strcpy(asset->path, optarg);
				break;

			default:
				usage();
				exit(1);
				break;
		}
	}

// Force compression
	if(vconfig_in->driver == CAPTURE_BUZ ||
		vconfig_out->driver == PLAYBACK_BUZ)
	{
		strcpy(asset->compression, QUICKTIME_MJPA);
	}
}

void MonteVideo::load_defaults()
{
	defaults = new Defaults("~/.montevideo");
	defaults->load();

	cpus = calculate_cpus();
	if(!aconfig_in) aconfig_in = new AudioInConfig;
	if(!vconfig_in) vconfig_in = new VideoInConfig;

	aconfig_in->load_defaults(defaults);
	vconfig_in->load_defaults(defaults);

	if(!aconfig_out) aconfig_out = new AudioOutConfig(PLAYBACK_LOCALHOST, 0);
	if(!vconfig_out) vconfig_out = new VideoOutConfig(PLAYBACK_LOCALHOST, 0);
	aconfig_out->load_defaults(defaults);
	vconfig_out->load_defaults(defaults);

	canvas_w = defaults->get("CANVAS_W", 720);
	canvas_h = defaults->get("CANVAS_H", 540);
	window_x = defaults->get("WINDOW_X", 0);
	window_y = defaults->get("WINDOW_Y", 0);
	field_order = defaults->get("FIELD_ORDER", 0);
	deinterlace = defaults->get("DEINTERLACE", 0);

	if(!asset) asset = new Asset;
	sprintf(asset->path, "montevideo");
	defaults->get("PATH", asset->path);
	sprintf(asset->compression, QUICKTIME_MJPA);
	defaults->get("COMPRESSION", asset->compression);
	asset->width = defaults->get("WIDTH", 720);
	asset->height = defaults->get("HEIGHT", 480);
	asset->rate = defaults->get("SAMPLE_RATE", 48000);
	asset->frame_rate = defaults->get("FRAME_RATE", 29.97);
	asset->quality = defaults->get("QUALITY", 80);
	asset->channels = defaults->get("CHANNELS", 2);
	state = defaults->get("STATE", state);
	current_channel = defaults->get("CURRENT_CHANNEL", 0);
	current_outchannel = defaults->get("CURRENT_OUTCHANNEL", 0);
	loop_duration = defaults->get("LOOP_DURATION", 1);
}

void MonteVideo::save_defaults()
{
	if(window)
	{
		window_x = window->get_x();
		window_y = window->get_y();
	}
	aconfig_in->save_defaults(defaults);
	vconfig_in->save_defaults(defaults);
	aconfig_out->save_defaults(defaults);
	vconfig_out->save_defaults(defaults);

	defaults->update("CANVAS_W", canvas_w);
	defaults->update("CANVAS_H", canvas_h);
	defaults->update("WINDOW_X", window_x);
	defaults->update("WINDOW_Y", window_y);
	defaults->update("DEINTERLACE", deinterlace);

	defaults->update("PATH", asset->path);
	defaults->update("COMPRESSION", asset->compression);
	defaults->update("WIDTH", asset->width);
	defaults->update("HEIGHT", asset->height);
	defaults->update("SAMPLE_RATE", asset->rate);
	defaults->update("FRAME_RATE", asset->frame_rate);
	defaults->update("QUALITY", asset->quality);
	defaults->update("CHANNELS", asset->channels);;
	defaults->update("STATE", state);
	defaults->update("CURRENT_CHANNEL", current_channel);
	defaults->update("CURRENT_OUTCHANNEL", current_outchannel);
	defaults->update("LOOP_DURATION", loop_duration);
	defaults->save();
}

int MonteVideo::calculate_cpus()
{
	FILE *proc;
	int result = 1;
	char string[1024];

	if(proc = fopen("/proc/cpuinfo", "r"))
	{
		while(!feof(proc))
		{
			fgets(string, 1024, proc);
			if(!strncasecmp(string, "processor", 9))
			{
				char *ptr = strchr(string, ':');
				if(ptr)
				{
					ptr++;
					result = atol(ptr) + 1;
				}
			}
		}
		fclose(proc);
	}
	return result;
}

ArrayList<Channel*>* MonteVideo::get_current_channeldb()
{
	switch(vconfig_in->driver)
	{
		case VIDEO4LINUX:
			return &channeldb_v4l;
			break;
		case SCREENCAPTURE:
			return 0;
			break;
		case CAPTURE_BUZ:
			return &channeldb_buz;
			break;
	}
	return 0;
}


Channel* MonteVideo::get_current_channel()
{
	if(!get_current_channeldb()) return 0;
	if(current_channel >= 0 && current_channel < get_current_channeldb()->total)
	{
		return get_current_channeldb()->values[current_channel];
	}
	return 0;
}

Channel* MonteVideo::get_current_outchannel()
{
	if(current_outchannel >= 0 && current_outchannel < channeldb_buz.total)
	{
		return channeldb_buz.values[current_outchannel];
	}
	return 0;
}

void MonteVideo::next_channel()
{
//printf("MonteVideo::next_channel 1\n");
	if(current_channel < get_current_channeldb()->total)
		current_channel++;
	else
		current_channel = 0;
//printf("MonteVideo::next_channel 1\n");
	Channel *channel = get_current_channel();
//printf("MonteVideo::next_channel 1\n");
	if(channel) video_in->set_channel(channel);
printf("MonteVideo::next_channel 1\n");
	interface->show_channel();
printf("MonteVideo::next_channel 2\n");
}

void MonteVideo::prev_channel()
{
	if(current_channel > 0)
		current_channel--;
	else
		current_channel = get_current_channeldb()->total - 1;
	Channel *channel = get_current_channel();
	if(channel) video_in->set_channel(channel);
	interface->show_channel();
}



void MonteVideo::open_input()
{
	if(!video_in) video_in = new VideoDevice;
	video_in->open_input(vconfig_in, 
		0, 
		0, 
		1,
		asset->frame_rate);
	Channel *channel = get_current_channel();
	if(channel) video_in->set_channel(channel);

//printf("MonteVideo::open_input %d\n", current_channel);

	if(!audio_in) audio_in = new AudioDevice;
	audio_in->open_input(aconfig_in, asset->rate, AUDIOBUFFER);
}

void MonteVideo::open_output()
{
	if(!video_out) video_out = new VideoDevice;
	video_out->open_output(vconfig_out,
		asset->frame_rate,
		asset->width,
		asset->height,
		window ? window->canvas : 0);
printf("MonteVideo::open_output 1 %p\n", get_current_outchannel());
	video_out->set_channel(get_current_outchannel());
	video_out->set_quality(asset->quality);
	video_out->set_cpus(cpus);

	if(!audio_out) audio_out = new AudioDevice;
	audio_out->open_output(aconfig_out, 
		asset->rate, 
		AUDIOBUFFER);
	audio_out->start_playback();
}

int MonteVideo::calculate_input_size()
{
	long fragment_size = 1;
	while(fragment_size < asset->rate / RECORD_SPEED) fragment_size *= 2;
	fragment_size /= 2;
	if(fragment_size > AUDIOBUFFER) fragment_size = AUDIOBUFFER;
	return fragment_size;
}

long MonteVideo::input_sync_position()
{
	switch(state)
	{
		case LOOPBACK:
			return record_audio->get_sync_position();
			break;
	}
}

long MonteVideo::output_sync_position()
{
	switch(state)
	{
		case LOOPBACK:
			return play_audio->get_sync_position();
			break;
	}
}

void MonteVideo::close_devices()
{
}

void MonteVideo::open_loopback_data()
{
	database->loopback->open_file(0);
}

void MonteVideo::write_audio(float **buffer, long samples, long time)
{
	switch(state)
	{
		case LOOPBACK:
			database->loopback->write_audio(buffer, samples, time);
			break;
	}
}

void MonteVideo::write_video(VFrame *frame, long time)
{
	switch(state)
	{
		case LOOPBACK:
			database->loopback->write_video(frame, time);
			break;
	}
}


int MonteVideo::read_audio(float **buffer, long samples, long time)
{
	switch(state)
	{
		case LOOPBACK:
			return database->loopback->read_audio(buffer, samples, time);
			break;
	}
}

int MonteVideo::read_video(VFrame *frame, long &time)
{
	int result = 1;
	switch(state)
	{
		case LOOPBACK:
			result = database->loopback->read_video(frame, time);
//printf("MonteVideo::read_video 1 %d\n", frame->get_compressed_size());
			break;
	}
	return result;
}

void MonteVideo::start_recording()
{
	record_audio = new MVAudioIn(this);
	record_video = new MVVideoIn(this);
	record_audio->start();
	record_video->start();
}

void MonteVideo::start_playback()
{
	play_audio = new MVAudioOut(this);
	play_video = new MVVideoOut(this);
	play_audio->start();
	play_video->start();
}

void MonteVideo::wait_recording()
{
	record_audio->join();
	record_video->join();
	delete record_audio;
	delete record_video;
}

void MonteVideo::wait_playback()
{
	play_audio->join();
	play_video->join();
	delete play_audio;
	delete play_video;
}

void MonteVideo::stop_all()
{
	switch(state)
	{
		case LOOPBACK:
			done = 1;
			play_audio->done = 1;
			play_video->done = 1;
			record_audio->done = 1;
			record_video->done = 1;
			break;
	}
}

void MonteVideo::init_interface()
{
	interface = new MVInterface(this);
	switch(vconfig_out->driver)
	{
		case PLAYBACK_X11:
			window = new MVWindow(this);
			window->create_objects();
			break;
		case PLAYBACK_BUZ:
			console = new MVConsole(this);
			break;
	}
	interface->start();
}

void MonteVideo::create_objects(int argc, char *argv[])
{
//printf("MonteVideo::create_objects 1\n");
	load_defaults();
	parse_args(argc, argv);
	save_defaults();

//printf("MonteVideo::create_objects 1\n");
	database = new MVDatabase(this);
//printf("MonteVideo::create_objects 1\n");
	database->load();
//printf("MonteVideo::create_objects 1\n");
	init_interface();
//printf("MonteVideo::create_objects 1\n");
	MWindow::init_tuner(channeldb_v4l, "channels_v4l");
	MWindow::init_tuner(channeldb_buz, "channels_buz");
//printf("MonteVideo::create_objects 1\n");
	overlay = new VFrame(0, 
		asset->width,
		asset->height,
		BC_YUVA8888,
		-1);
//printf("MonteVideo::create_objects 2\n");
}

// Display composited frame
void MonteVideo::write_composite()
{
}

void MonteVideo::do_loopback()
{
// Playback threads
	play_audio = new MVAudioOut(this);
	play_video = new MVVideoOut(this);

//printf("MonteVideo::do_loopback 1\n");
	open_input();
//printf("MonteVideo::do_loopback 1\n");
	open_output();
//printf("MonteVideo::do_loopback 1\n");
	open_loopback_data();
//printf("MonteVideo::do_loopback 1\n");
	start_recording();
// Give the loop time to fill
//	Timer timer;
//	timer.delay((long)((float)AUDIOBUFFER * 2 / asset->rate * 1000));
//printf("MonteVideo::do_loopback 1\n");
	start_playback();
//printf("MonteVideo::do_loopback 1\n");
	wait_recording();
//printf("MonteVideo::do_loopback 1\n");
	wait_playback();
//printf("MonteVideo::do_loopback 2\n");

	delete play_audio;
	delete play_video;
}

void MonteVideo::do_inactive()
{
	write_composite();
}


void MonteVideo::main_loop()
{
	while(!done)
	{
		switch(state)
		{
			case LOOPBACK:
				do_loopback();
				break;

			case INACTIVE:
				do_inactive();
				break;
		}
		save_defaults();
	}
}


