#include "filexml.h"
#include "blur.h"
#include "blurwindow.h"
#include "defaults.h"
#include "keyframe.h"
#include "picon_png.h"
#include "vframe.h"

#include <math.h>
#include <string.h>




PluginClient* new_plugin(PluginServer *server)
{
	return new BlurMain(server);
}





BlurMain::BlurMain(PluginServer *server)
 : PluginVClient(server)
{
	vertical = 1;
	horizontal = 1;
	radius = 5;
	defaults = 0;
	thread = 0;
	load_defaults();
}

BlurMain::~BlurMain()
{
	if(thread)
	{
// Set result to 0 to indicate a server side close
		thread->window->set_done(0);
		thread->completion.lock();
		delete thread;
	}

	save_defaults();
	delete defaults;
}

char* BlurMain::plugin_title() { return "Blur"; }
int BlurMain::is_realtime() { return 1; }

VFrame* BlurMain::new_picon()
{
	return new VFrame(picon_png);
}


int BlurMain::start_realtime()
{
	int y1, y2, y_increment;
	last_radius = radius;
	redo_buffers = 1;

	y_increment = project_frame_h / (smp + 1);
	y1 = 0;

	engine = new BlurEngine*[(smp + 1)];
	for(int i = 0; i < (smp + 1); i++)
	{
		y2 = y1 + y_increment;
		if(i == (smp + 1) - 1 && y2 < project_frame_h - 1) y2 = project_frame_h - 1;
		engine[i] = new BlurEngine(this, y1, y2);
		engine[i]->start();
		y1 += y_increment;
	}
	return 0;
}

int BlurMain::stop_realtime()
{
	for(int i = 0; i < (smp + 1); i++)
		delete engine[i];
	delete [] engine;
	return 0;
}

int BlurMain::process_realtime(VFrame *input_ptr, VFrame *output_ptr)
{
	int i, j, k, l;
	unsigned char **input_rows, **output_rows;

	load_configuration();

	for(i = 0; i < (smp + 1); i++)
		engine[i]->reconfigure();

	input_rows = input_ptr->get_rows();
	output_rows = output_ptr->get_rows();

	if(radius < 2 || (!vertical && !horizontal))
	{
// Data never processed so copy if necessary
//printf("BlurMain::process_realtime 2 %d\n", radius);
		if(input_rows[0] != output_rows[0])
		{
			output_ptr->copy_from(input_ptr);
		}
	}
	else
	{
// Process blur
//printf("BlurMain::process_realtime 3 %d\n", (smp + 1));
		for(i = 0; i < (smp + 1); i++)
		{
			engine[i]->start_process_frame(output_ptr, input_ptr);
		}

		for(i = 0; i < (smp + 1); i++)
		{
			engine[i]->wait_process_frame();
		}
	}

	return 0;
}

int BlurMain::show_gui()
{
	load_configuration();
	thread = new BlurThread(this);
	thread->start();
	return 0;
}

int BlurMain::set_string()
{
	if(thread) thread->window->set_title(gui_string);
	return 0;
}

void BlurMain::raise_window()
{
	if(thread)
	{
		thread->window->raise_window();
		thread->window->flush();
	}
}

int BlurMain::load_defaults()
{
	char directory[1024], string[1024];
// set the default directory
	sprintf(directory, "%sblur.rc", BCASTDIR);

// load the defaults
	defaults = new Defaults(directory);
	defaults->load();

	vertical = defaults->get("VERTICAL", vertical);
	horizontal = defaults->get("HORIZONTAL", horizontal);
	radius = defaults->get("RADIUS", radius);
	return 0;
}

int BlurMain::save_defaults()
{
	defaults->update("VERTICAL", vertical);
	defaults->update("HORIZONTAL", horizontal);
	defaults->update("RADIUS", radius);
	defaults->save();
	return 0;
}

void BlurMain::load_configuration()
{
	KeyFrame *prev_keyframe, *next_keyframe;
//printf("BlurMain::load_configuration 1\n");

	prev_keyframe = get_prev_keyframe(-1);
	next_keyframe = get_next_keyframe(-1);
// Must also switch between interpolation between keyframes and using first keyframe
//printf("BlurMain::load_configuration %s\n", prev_keyframe->data);
	read_data(prev_keyframe);
}

void BlurMain::save_data(KeyFrame *keyframe)
{
	FileXML output;

// cause data to be stored directly in text
	output.set_shared_string(keyframe->data, MESSAGESIZE);
	output.tag.set_title("BLUR");
	output.tag.set_property("VERTICAL", vertical);
	output.tag.set_property("HORIZONTAL", horizontal);
	output.tag.set_property("RADIUS", radius);
	output.append_tag();
	output.terminate_string();
}

void BlurMain::read_data(KeyFrame *keyframe)
{
	FileXML input;

	input.set_shared_string(keyframe->data, strlen(keyframe->data));

	int result = 0;

	while(!result)
	{
		result = input.read_tag();

		if(!result)
		{
			if(input.tag.title_is("BLUR"))
			{
				vertical = input.tag.get_property("VERTICAL", vertical);
				horizontal = input.tag.get_property("HORIZONTAL", horizontal);
				radius = input.tag.get_property("RADIUS", radius);
			}
		}
	}
	if(thread) 
	{
		thread->window->vertical->update(vertical);
		thread->window->horizontal->update(horizontal);
		thread->window->radius->update(radius);
	}
	redo_buffers = 1;
}









BlurEngine::BlurEngine(BlurMain *plugin, int start_out, int end_out)
 : Thread()
{
	int size = plugin->project_frame_w > plugin->project_frame_h ? plugin->project_frame_w : plugin->project_frame_h;
	this->plugin = plugin;
	this->start_out = start_out;
	this->end_out = end_out;
	last_frame = 0;
	val_p = new pixel_f[size];
	val_m = new pixel_f[size];
	src = new pixel_f[size];
	dst = new pixel_f[size];
	set_synchronous(1);
	input_lock.lock();
	output_lock.lock();
}

BlurEngine::~BlurEngine()
{
	last_frame = 1;
	input_lock.unlock();
	join();
}

int BlurEngine::start_process_frame(VFrame *output, VFrame *input)
{
	this->output = output;
	this->input = input;
	input_lock.unlock();
	return 0;
}

int BlurEngine::wait_process_frame()
{
	output_lock.lock();
	return 0;
}

void BlurEngine::run()
{
	int i, j, k, l;
	unsigned char **input_rows, **output_rows;
	int strip_size;

	while(1)
	{
		input_lock.lock();
		if(last_frame)
		{
			output_lock.unlock();
			return;
		}

		start_in = start_out - plugin->radius;
		end_in = end_out + plugin->radius;
		if(start_in < 0) start_in = 0;
		if(end_in > plugin->project_frame_h) end_in = plugin->project_frame_h;
		strip_size = end_in - start_in;
		input_rows = input->get_rows();
		output_rows = output->get_rows();
		color_model = input->get_color_model();

		switch(color_model)
		{
			case BC_RGB888:
			case BC_YUV888:
				vmax = 0xff;
				if(plugin->vertical)
				{
// Vertical pass
					for(j = 0; j < plugin->project_frame_w; j++)
					{
						bzero(val_p, sizeof(pixel_f) * (end_in - start_in));
						bzero(val_m, sizeof(pixel_f) * (end_in - start_in));
						for(l = 0, k = start_in; k < end_in; l++, k++)
						{
							src[l].r = (float)input_rows[k][j * 3];
							src[l].g = (float)input_rows[k][j * 3 + 1];
							src[l].b = (float)input_rows[k][j * 3 + 2];
							src[l].a = vmax;
						}

						blur_strip(j, strip_size);

						for(l = start_out - start_in, k = start_out; k < end_out; l++, k++)
						{
							output_rows[k][j * 3] = (unsigned char)dst[l].r;
							output_rows[k][j * 3 + 1] = (unsigned char)dst[l].g;
							output_rows[k][j * 3 + 2] = (unsigned char)dst[l].b;
						}
					}
				}

				if(plugin->horizontal)
				{
// Horizontal pass
					if(plugin->vertical) input_rows = output_rows;

					for(j = start_out; j < end_out; j++)
					{
						bzero(val_p, sizeof(pixel_f) * plugin->project_frame_w);
						bzero(val_m, sizeof(pixel_f) * plugin->project_frame_w);
						for(k = 0; k < plugin->project_frame_w; k++)
						{
							src[k].r = (float)input_rows[j][k * 3];
							src[k].g = (float)input_rows[j][k * 3 + 1];
							src[k].b = (float)input_rows[j][k * 3 + 2];
							src[k].a = vmax;
						}

						blur_strip(j, plugin->project_frame_w);

						for(k = 0; k < plugin->project_frame_w; k++)
						{
							output_rows[j][k * 3] = (unsigned char)dst[k].r;
							output_rows[j][k * 3 + 1] = (unsigned char)dst[k].g;
							output_rows[j][k * 3 + 2] = (unsigned char)dst[k].b;
						}
					}
				}

//printf("Blur:: %d %d\n", plugin->vertical, plugin->horizontal);

// for(i = 0; i < 2000; i++)
// {
// 	output_rows[0][i] = 255;
// }
				break;
			case BC_RGBA8888:
			case BC_YUVA8888:
				vmax = 0xff;
				if(plugin->vertical)
				{
// Vertical pass
					for(j = 0; j < plugin->project_frame_w; j++)
					{
						bzero(val_p, sizeof(pixel_f) * (end_in - start_in));
						bzero(val_m, sizeof(pixel_f) * (end_in - start_in));
						for(l = 0, k = start_in; k < end_in; l++, k++)
						{
							src[l].r = (float)input_rows[k][j * 4];
							src[l].g = (float)input_rows[k][j * 4 + 1];
							src[l].b = (float)input_rows[k][j * 4 + 2];
							src[l].a = (float)input_rows[k][j * 4 + 3];
						}

						blur_strip(j, strip_size);

						for(l = start_out - start_in, k = start_out; k < end_out; l++, k++)
						{
							output_rows[k][j * 4] = (unsigned char)dst[l].r;
							output_rows[k][j * 4 + 1] = (unsigned char)dst[l].g;
							output_rows[k][j * 4 + 2] = (unsigned char)dst[l].b;
							output_rows[k][j * 4 + 3] = (unsigned char)dst[l].a;
						}
					}
				}

				if(plugin->horizontal)
				{
// Horizontal pass
					if(plugin->vertical) input_rows = output_rows;

					for(j = start_out; j < end_out; j++)
					{
						bzero(val_p, sizeof(pixel_f) * plugin->project_frame_w);
						bzero(val_m, sizeof(pixel_f) * plugin->project_frame_w);
						for(k = 0; k < plugin->project_frame_w; k++)
						{
							src[k].r = (float)input_rows[j][k * 4];
							src[k].g = (float)input_rows[j][k * 4 + 1];
							src[k].b = (float)input_rows[j][k * 4 + 2];
							src[k].a = (float)input_rows[j][k * 4 + 3];
						}

						blur_strip(j, plugin->project_frame_w);

						for(k = 0; k < plugin->project_frame_w; k++)
						{
							output_rows[j][k * 4] = (unsigned char)dst[k].r;
							output_rows[j][k * 4 + 1] = (unsigned char)dst[k].g;
							output_rows[j][k * 4 + 2] = (unsigned char)dst[k].b;
							output_rows[j][k * 4 + 3] = (unsigned char)dst[k].a;
						}
					}
				}
				break;
			case BC_RGB161616:
			case BC_YUV161616:
				vmax = 0xffff;
				if(plugin->vertical)
				{
// Vertical pass
					for(j = 0; j < plugin->project_frame_w; j++)
					{
						bzero(val_p, sizeof(pixel_f) * (end_in - start_in));
						bzero(val_m, sizeof(pixel_f) * (end_in - start_in));
						for(l = 0, k = start_in; k < end_in; l++, k++)
						{
							src[l].r = (float)((u_int16_t*)input_rows[k])[j * 3];
							src[l].g = (float)((u_int16_t*)input_rows[k])[j * 3 + 1];
							src[l].b = (float)((u_int16_t*)input_rows[k])[j * 3 + 2];
							src[l].a = vmax;
						}

						blur_strip(j, strip_size);

						for(l = start_out - start_in, k = start_out; k < end_out; l++, k++)
						{
							((u_int16_t*)output_rows[k])[j * 3] = (u_int16_t)dst[l].r;
							((u_int16_t*)output_rows[k])[j * 3 + 1] = (u_int16_t)dst[l].g;
							((u_int16_t*)output_rows[k])[j * 3 + 2] = (u_int16_t)dst[l].b;
						}
					}
				}

				if(plugin->horizontal)
				{
// Horizontal pass
					if(plugin->vertical) input_rows = output_rows;

					for(j = start_out; j < end_out; j++)
					{
						bzero(val_p, sizeof(pixel_f) * plugin->project_frame_w);
						bzero(val_m, sizeof(pixel_f) * plugin->project_frame_w);
						for(k = 0; k < plugin->project_frame_w; k++)
						{
							src[k].r = (float)((u_int16_t*)input_rows[j])[k * 3];
							src[k].g = (float)((u_int16_t*)input_rows[j])[k * 3 + 1];
							src[k].b = (float)((u_int16_t*)input_rows[j])[k * 3 + 2];
							src[k].a = vmax;
						}

						blur_strip(j, plugin->project_frame_w);

						for(k = 0; k < plugin->project_frame_w; k++)
						{
							((u_int16_t*)output_rows[j])[k * 3] = (u_int16_t)dst[k].r;
							((u_int16_t*)output_rows[j])[k * 3 + 1] = (u_int16_t)dst[k].g;
							((u_int16_t*)output_rows[j])[k * 3 + 2] = (u_int16_t)dst[k].b;
						}
					}
				}
				break;
			case BC_RGBA16161616:
			case BC_YUVA16161616:
				vmax = 0xffff;
				if(plugin->vertical)
				{
// Vertical pass
					for(j = 0; j < plugin->project_frame_w; j++)
					{
						bzero(val_p, sizeof(pixel_f) * (end_in - start_in));
						bzero(val_m, sizeof(pixel_f) * (end_in - start_in));
						for(l = 0, k = start_in; k < end_in; l++, k++)
						{
							src[l].r = (float)((u_int16_t*)input_rows[k])[j * 4];
							src[l].g = (float)((u_int16_t*)input_rows[k])[j * 4 + 1];
							src[l].b = (float)((u_int16_t*)input_rows[k])[j * 4 + 2];
							src[l].a = (float)((u_int16_t*)input_rows[k])[j * 4 + 3];
						}

						blur_strip(j, strip_size);

						for(l = start_out - start_in, k = start_out; k < end_out; l++, k++)
						{
							((u_int16_t*)output_rows[k])[j * 4] = (u_int16_t)dst[l].r;
							((u_int16_t*)output_rows[k])[j * 4 + 1] = (u_int16_t)dst[l].g;
							((u_int16_t*)output_rows[k])[j * 4 + 2] = (u_int16_t)dst[l].b;
							((u_int16_t*)output_rows[k])[j * 4 + 3] = (u_int16_t)dst[l].a;
						}
					}
				}

				if(plugin->horizontal)
				{
// Horizontal pass
					if(plugin->vertical) input_rows = output_rows;

					for(j = start_out; j < end_out; j++)
					{
						bzero(val_p, sizeof(pixel_f) * plugin->project_frame_w);
						bzero(val_m, sizeof(pixel_f) * plugin->project_frame_w);
						for(k = 0; k < plugin->project_frame_w; k++)
						{
							src[k].r = (float)((u_int16_t*)input_rows[j])[k * 4];
							src[k].g = (float)((u_int16_t*)input_rows[j])[k * 4 + 1];
							src[k].b = (float)((u_int16_t*)input_rows[j])[k * 4 + 2];
							src[k].a = (float)((u_int16_t*)input_rows[j])[k * 4 + 3];
						}

						blur_strip(j, plugin->project_frame_w);

						for(k = 0; k < plugin->project_frame_w; k++)
						{
							((u_int16_t*)output_rows[j])[k * 4] = (u_int16_t)dst[k].r;
							((u_int16_t*)output_rows[j])[k * 4 + 1] = (u_int16_t)dst[k].g;
							((u_int16_t*)output_rows[j])[k * 4 + 2] = (u_int16_t)dst[k].b;
							((u_int16_t*)output_rows[j])[k * 4 + 3] = (u_int16_t)dst[k].a;
						}
					}
				}
				break;
		}

		output_lock.unlock();
	}
}

int BlurEngine::reconfigure()
{
	std_dev = sqrt(-(double)(plugin->radius * plugin->radius) / (2 * log (1.0 / 255.0)));
	get_constants();
}

int BlurEngine::get_constants()
{
	int i;
	double constants[8];
	double div;

	div = sqrt(2 * M_PI) * std_dev;
	constants[0] = -1.783 / std_dev;
	constants[1] = -1.723 / std_dev;
	constants[2] = 0.6318 / std_dev;
	constants[3] = 1.997  / std_dev;
	constants[4] = 1.6803 / div;
	constants[5] = 3.735 / div;
	constants[6] = -0.6803 / div;
	constants[7] = -0.2598 / div;

	n_p[0] = constants[4] + constants[6];
	n_p[1] = exp(constants[1]) *
				(constants[7] * sin(constants[3]) -
				(constants[6] + 2 * constants[4]) * cos(constants[3])) +
				exp(constants[0]) *
				(constants[5] * sin(constants[2]) -
				(2 * constants[6] + constants[4]) * cos(constants[2]));

	n_p[2] = 2 * exp(constants[0] + constants[1]) *
				((constants[4] + constants[6]) * cos(constants[3]) * 
				cos(constants[2]) - constants[5] * 
				cos(constants[3]) * sin(constants[2]) -
				constants[7] * cos(constants[2]) * sin(constants[3])) +
				constants[6] * exp(2 * constants[0]) +
				constants[4] * exp(2 * constants[1]);

	n_p[3] = exp(constants[1] + 2 * constants[0]) *
				(constants[7] * sin(constants[3]) - 
				constants[6] * cos(constants[3])) +
				exp(constants[0] + 2 * constants[1]) *
				(constants[5] * sin(constants[2]) - constants[4] * 
				cos(constants[2]));
	n_p[4] = 0.0;

	d_p[0] = 0.0;
	d_p[1] = -2 * exp(constants[1]) * cos(constants[3]) -
				2 * exp(constants[0]) * cos(constants[2]);

	d_p[2] = 4 * cos(constants[3]) * cos(constants[2]) * 
				exp(constants[0] + constants[1]) +
				exp(2 * constants[1]) + exp (2 * constants[0]);

	d_p[3] = -2 * cos(constants[2]) * exp(constants[0] + 2 * constants[1]) -
				2 * cos(constants[3]) * exp(constants[1] + 2 * constants[0]);

	d_p[4] = exp(2 * constants[0] + 2 * constants[1]);

	for(i = 0; i < 5; i++) d_m[i] = d_p[i];

	n_m[0] = 0.0;
	for(i = 1; i <= 4; i++)
		n_m[i] = n_p[i] - d_p[i] * n_p[0];

	double sum_n_p, sum_n_m, sum_d;
	double a, b;

	sum_n_p = 0.0;
	sum_n_m = 0.0;
	sum_d = 0.0;
	for(i = 0; i < 5; i++)
	{
		sum_n_p += n_p[i];
		sum_n_m += n_m[i];
		sum_d += d_p[i];
	}

	a = sum_n_p / (1 + sum_d);
	b = sum_n_m / (1 + sum_d);

	for (i = 0; i < 5; i++)
	{
		bd_p[i] = d_p[i] * a;
		bd_m[i] = d_m[i] * b;
	}
	return 0;
}

#define BOUNDARY(x) if((x) > vmax) (x) = vmax; else if((x) < 0) (x) = 0;

int BlurEngine::transfer_pixels(pixel_f *src1, pixel_f *src2, pixel_f *dest, int size)
{
	register int i;
	register float sum;

	for(i = 0; i < size; i++)
    {
		sum = src1[i].r + src2[i].r;
		BOUNDARY(sum);
		dest[i].r = sum;
		sum = src1[i].g + src2[i].g;
		BOUNDARY(sum);
		dest[i].g = sum;
		sum = src1[i].b + src2[i].b;
		BOUNDARY(sum);
		dest[i].b = sum;
		sum = src1[i].a + src2[i].a;
		BOUNDARY(sum);
		dest[i].a = sum;
    }
	return 0;
}


int BlurEngine::multiply_alpha(pixel_f *row, int size)
{
	register int i;
	register float alpha;

	for(i = 0; i < size; i++)
	{
		alpha = (float)row[i].a / vmax;
		row[i].r *= alpha;
		row[i].g *= alpha;
		row[i].b *= alpha;
	}
	return 0;
}

int BlurEngine::separate_alpha(pixel_f *row, int size)
{
	register int i;
	register float alpha;
	register float result;
	
	for(i = 0; i < size; i++)
	{
		if(row[i].a > 0 && row[i].a < vmax)
		{
			alpha = (float)row[i].a / vmax;
			result = (float)row[i].r / alpha;
			row[i].r = (result > vmax ? vmax : result);
			result = (float)row[i].g / alpha;
			row[i].g = (result > vmax ? vmax : result);
			result = (float)row[i].b / alpha;
			row[i].b = (result > vmax ? vmax : result);
		}
	}
	return 0;
}

int BlurEngine::blur_strip(int &j, int &size)
{
	multiply_alpha(src, size);

	sp_p = src;
	sp_m = src + size - 1;
	vp = val_p;
	vm = val_m + size - 1;

	initial_p = sp_p[0];
	initial_m = sp_m[0];

	register int l;
	for(register int k = 0; k < size; k++)
	{
		terms = (k < 4) ? k : 4;
		for(l = 0; l <= terms; l++)
		{
			vp->r += n_p[l] * sp_p[-l].r - d_p[l] * vp[-l].r;
			vm->r += n_m[l] * sp_m[l].r - d_m[l] * vm[l].r;
			vp->g += n_p[l] * sp_p[-l].g - d_p[l] * vp[-l].g;
			vm->g += n_m[l] * sp_m[l].g - d_m[l] * vm[l].g;
			vp->b += n_p[l] * sp_p[-l].b - d_p[l] * vp[-l].b;
			vm->b += n_m[l] * sp_m[l].b - d_m[l] * vm[l].b;
			vp->a += n_p[l] * sp_p[-l].a - d_p[l] * vp[-l].a;
			vm->a += n_m[l] * sp_m[l].a - d_m[l] * vm[l].a;
		}
		for( ; l <= 4; l++)
		{
			vp->r += (n_p[l] - bd_p[l]) * initial_p.r;
			vm->r += (n_m[l] - bd_m[l]) * initial_m.r;
			vp->g += (n_p[l] - bd_p[l]) * initial_p.g;
			vm->g += (n_m[l] - bd_m[l]) * initial_m.g;
			vp->b += (n_p[l] - bd_p[l]) * initial_p.b;
			vm->b += (n_m[l] - bd_m[l]) * initial_m.b;
			vp->a += (n_p[l] - bd_p[l]) * initial_p.a;
			vm->a += (n_m[l] - bd_m[l]) * initial_m.a;
		}
		sp_p++;
		sp_m--;
		vp++;
		vm--;
	}
	transfer_pixels(val_p, val_m, dst, size);
	separate_alpha(dst, size);
	return 0;
}

