#include "colormodels.h"
#include "filexml.h"
#include "filesystem.h"
#include "picon_png.h"
#include "title.h"
#include "titlewindow.h"

#include <stdio.h>
#include <string.h>

PluginClient* new_plugin(PluginServer *server)
{
	return new TitleMain(server);
}


TitleConfig::TitleConfig()
{
	style = 0;
	color = BLACK;
	size = 24;
	motion_strategy = NO_MOTION;
	justification = JUSTIFY_LEFT;
	fade_in = 0.0;
	fade_out = 0.0;
	sprintf(font, "fixed");
	text[0] = 0;
}


TitleMain::TitleMain(PluginServer *server)
 : PluginVClient(server)
{
	thread = 0;
	load_defaults();
}

TitleMain::~TitleMain()
{
//printf("TitleMain::~TitleMain 1\n");
	if(thread)
	{
// Set result to 0 to indicate a server side close
		thread->window->set_done(0);
		thread->completion.lock();
		delete thread;
	}

//printf("TitleMain::~TitleMain 1\n");
	save_defaults();
	delete defaults;
//printf("TitleMain::~TitleMain 2\n");
}

char* TitleMain::plugin_title() { return "Title"; }
int TitleMain::is_realtime() { return 1; }

VFrame* TitleMain::new_picon()
{
	return new VFrame(picon_png);
}


int TitleMain::start_realtime()
{
	return 0;
}

int TitleMain::stop_realtime()
{
	return 0;
}

int TitleMain::process_realtime(VFrame *input_ptr, VFrame *output_ptr)
{
	load_configuration();
	return 0;
}

int TitleMain::show_gui()
{
	load_configuration();
	thread = new TitleThread(this);
	thread->start();
	return 0;
}

int TitleMain::set_string()
{
	if(thread) thread->window->set_title(gui_string);
	return 0;
}

void TitleMain::raise_window()
{
	if(thread)
	{
		thread->window->raise_window();
		thread->window->flush();
	}
}

int TitleMain::load_defaults()
{
	char directory[1024], text_path[1024];
// set the default directory
	sprintf(directory, "%stitle.rc", BCASTDIR);

// load the defaults
	defaults = new Defaults(directory);
	defaults->load();

	defaults->get("FONT", config.font);
	config.style = defaults->get("STYLE", (long)config.style);
	config.size = defaults->get("SIZE", config.size);
	config.color = defaults->get("COLOR", config.color);
	config.motion_strategy = defaults->get("MOTION_STRATEGY", config.motion_strategy);
	config.justification = defaults->get("JUSTIFICATION", config.justification);
	config.fade_in = defaults->get("FADE_IN", config.fade_in);
	config.fade_out = defaults->get("FADE_OUT", config.fade_out);
	config.x = defaults->get("TITLE_X", config.x);
	config.y = defaults->get("TITLE_Y", config.y);
	window_w = defaults->get("WINDOW_W", 640);
	window_h = defaults->get("WINDOW_H", 480);

// Store text in separate path to isolate special characters
	FileSystem fs;
	sprintf(text_path, "%stitle_text.rc", BCASTDIR);
	fs.complete_path(text_path);
	FILE *fd = fopen(text_path, "rb");
	if(fd)
	{
		fseek(fd, 0, SEEK_END);
		long len = ftell(fd);
		fseek(fd, 0, SEEK_SET);
		fread(config.text, len, 1, fd);
		config.text[len] = 0;
//printf("TitleMain::load_defaults %s\n", config.text);
		fclose(fd);
	}
	else
		config.text[0] = 0;
	return 0;
}

int TitleMain::save_defaults()
{
	char text_path[1024];

	defaults->update("FONT", config.font);
	defaults->update("STYLE", (long)config.style);
	defaults->update("SIZE", config.size);
	defaults->update("COLOR", config.color);
	defaults->update("MOTION_STRATEGY", config.motion_strategy);
	defaults->update("JUSTIFICATION", config.justification);
	defaults->update("FADE_IN", config.fade_in);
	defaults->update("FADE_OUT", config.fade_out);
	defaults->update("TITLE_X", config.x);
	defaults->update("TITLE_Y", config.y);
	defaults->update("WINDOW_W", window_w);
	defaults->update("WINDOW_H", window_h);
	defaults->save();

// Store text in separate path to isolate special characters
	FileSystem fs;
	sprintf(text_path, "%stitle_text.rc", BCASTDIR);
	fs.complete_path(text_path);
	FILE *fd = fopen(text_path, "wb");
	if(fd)
	{
		fwrite(config.text, strlen(config.text), 1, fd);
		fclose(fd);
	}
	else
		perror("TitleMain::save_defaults");
	return 0;
}

void TitleMain::load_configuration()
{
	KeyFrame *prev_keyframe = get_prev_keyframe(-1);
	read_data(prev_keyframe);
}


void TitleMain::save_data(KeyFrame *keyframe)
{
	FileXML output;

// cause data to be stored directly in text
	output.set_shared_string(keyframe->data, MESSAGESIZE);
	output.tag.set_title("TITLE");
	output.tag.set_property("FONT", config.font);
	output.tag.set_property("STYLE", (long)config.style);
	output.tag.set_property("SIZE", config.size);
	output.tag.set_property("COLOR", config.color);
	output.tag.set_property("MOTION_STRATEGY", config.motion_strategy);
	output.tag.set_property("JUSTIFICATION", config.justification);
	output.tag.set_property("FADE_IN", config.fade_in);
	output.tag.set_property("FADE_OUT", config.fade_out);
	output.tag.set_property("TITLE_X", config.x);
	output.tag.set_property("TITLE_Y", config.y);
	output.append_tag();
	output.append_newline();
	
	output.append_text(config.text);

	output.tag.set_title("/TITLE");
	output.append_tag();
	output.append_newline();
	output.terminate_string();
//printf("TitleMain::save_data 1\n%s\n", output.string);
//printf("TitleMain::save_data 2\n%s\n", config.text);
}

void TitleMain::read_data(KeyFrame *keyframe)
{
	FileXML input;

	input.set_shared_string(keyframe->data, strlen(keyframe->data));

	int result = 0;
	int new_interlace = 0;
	int new_horizontal = 0;
	int new_luminance = 0;

	while(!result)
	{
		result = input.read_tag();

		if(!result)
		{
			if(input.tag.title_is("TITLE"))
			{
				input.tag.get_property("FONT", config.font);
				config.style = input.tag.get_property("STYLE", (long)config.style);
				config.size = input.tag.get_property("SIZE", config.size);
				config.color = input.tag.get_property("COLOR", config.color);
				config.motion_strategy = input.tag.get_property("MOTION_STRATEGY", config.motion_strategy);
				config.justification = input.tag.get_property("JUSTIFICATION", config.justification);
				config.fade_in = input.tag.get_property("FADE_IN", config.fade_in);
				config.fade_out = input.tag.get_property("FADE_OUT", config.fade_out);
				config.x = input.tag.get_property("TITLE_X", config.x);
				config.y = input.tag.get_property("TITLE_Y", config.y);
				strcpy(config.text, input.read_text());
//printf("TitleMain::read_data 1\n%s\n", input.string);
//printf("TitleMain::read_data 2\n%s\n", config.text);
			}
			else
			if(input.tag.title_is("/TITLE"))
			{
				result = 1;
			}
		}
	}

	if(thread) 
	{
		thread->window->update();
	}
}

char* TitleMain::motion_to_text(int motion)
{
	switch(motion)
	{
		case NO_MOTION: return "No motion"; break;
		case BOTTOM_TO_TOP: return "Bottom to top"; break;
		case RIGHT_TO_LEFT: return "Right to left"; break;
	}
}

int TitleMain::text_to_motion(char *text)
{
	for(int i = 0; i < TOTAL_PATHS; i++)
	{
		if(!strcasecmp(motion_to_text(i), text)) return i;
	}
	return 0;
}
